import tensorflow as tf
import tensorflow_probability as tfp
from scrinet.analysis.frequencyseries_batch import FrequencySeries

from scrinet.analysis.timeseries_batch import TimeSeries
from scrinet.analysis import matchedfilter_batch as matchedfilter
import numpy as np
from scrinet.sample.sample_helpers import WaveformParams, DetectorParams, tf_generate_surrogate_at_detector
from typing import NamedTuple, Any
from scrinet.sample.priors import Cosine

tfd = tfp.distributions
tfb = tfp.bijectors


class GwPriors(NamedTuple):
    """
    Object for passing waveform params to functions

    Attributes:
        distance: distance in metre - shape = N,1
        m_tot: total mass in Msun - shape = N,1
        mass_ratio: mass ratio (defined as q>1) - shape = N,1
        chi1: dimensionless spin for bh1  (defined over (-1, 1) - shape = N,1
        chi2: dimensionless spin for bh2 (defined over (-1, 1) - shape = N,1
        theta: sphharm polar - shape = N,1
        phi: sphharm azimuth - shape = N,1
        phase_shift: phase shift - shape = N,1
        ra: right ascension in radians - shape = N,1
        dec: declination in radians - shape = N,1
        psi: polarisation angle in radians - shape = N,1
        time: time (geocentrc maybe!?) - shape = N,1

    """
    # TODO Defaults are very bad defaults atm
    # TODO not sure of distribution type?

    distance: Any = tfd.Uniform(0.125, 1.0)
    m_tot: Any = tfd.Uniform(0.6, 1.)
    mass_ratio: Any = tfd.Uniform(0.125, 1.)
    chi_1: Any = tfd.Uniform(-1., 1.)
    chi_2: Any = tfd.Uniform(-1, 1.)
    theta: Any = tfd.Uniform(0, 1.)
    phi: Any = tfd.Uniform(0, 1.)
    phase_shift: Any = tfd.Uniform(0, 1.)
    ra: Any = tfd.Uniform(0, 1.)
    dec: Any = Cosine(minimum=-1, maximum=1)
    psi: Any = tfd.Uniform(0, 1.)
    time: Any = tfd.Uniform(0.99, 1.01)  # can make this very restrictive


def log_likelihood(
        signal,
        waveform_params,
        detector_params,
        n_batch,
        f_lower,
        delta_f,
        seglen,
        sample_rate,
        delta_t,
        amp_model,
        amp_basis,
        phase_model,
        phase_basis,
        psd,
        detector,
):
    template = tf_generate_surrogate_at_detector(
        waveform_params=waveform_params,
        detector_params=detector_params,
        n_batch=n_batch,
        f_lower=f_lower,
        seglen=seglen,
        sample_rate=sample_rate,
        amp_model=amp_model,
        amp_basis=amp_basis,
        phase_model=phase_model,
        phase_basis=phase_basis,
        detector=detector,
    )

    template = TimeSeries(template, delta_t=delta_t)

    template = FrequencySeries(template.to_frequencyseries().data, delta_f=delta_f)

    log_like = gw_log_like_of_template(
        signal, template, psd=psd, delta_f=delta_f, flow=f_lower, fhigh=3000
    )
    return log_like


# @tf.function(experimental_compile=True)
def gw_log_like_of_template(signal, template, delta_f=1., psd=None, flow=None, fhigh=None):
    """

    :param signal:
    :param template:
    :param delta_f:
    :param psd:
    :return:
    """

    duration = 1. / delta_f

    _, qtilde_fs_sig_temp, _ = matchedfilter.matched_filter_core(
        template, signal, psd=psd, low_frequency_cutoff=flow, high_frequency_cutoff=fhigh)
    # if time marginalised

    # d_inner_h = 4 / duration * \
    #             tf.math.reduce_sum(qtilde_fs_sig_temp.data, axis=1)

    d_inner_h = 4 / duration * \
                tf.signal.fft(qtilde_fs_sig_temp.data)

    _, qtilde_sig_sig, _ = matchedfilter.matched_filter_core(
        template, template, psd=psd, low_frequency_cutoff=flow, high_frequency_cutoff=fhigh)

    optimal_snr_squared = 4 / duration * \
                          tf.math.reduce_sum(qtilde_sig_sig.data, axis=1)

    # log_l = d_inner_h - optimal_snr_squared / 2
    # return tf.math.real(log_l)
    # MAY NEED TO CHECK WHETHER NOT MULTIPLYING BY PRIOR/DT MATTERS
    #also need to check when to add in opt_snr
    log_l = abs(d_inner_h) + tf.math.log(tf.math.bessel_i0e(abs(d_inner_h))) - \
     tf.reshape(tf.math.real(optimal_snr_squared / 2), (d_inner_h.shape[0], 1))
    # if time-marginalised
    log_l = tf.math.reduce_logsumexp(log_l, axis=1)
    return log_l


# @tf.function(experimental_compile=True, autograph=False, experimental_relax_shapes=True)
def simple_gw_prior(waveform_params: WaveformParams, detector_params: DetectorParams):
    prior_val = tf.zeros_like(waveform_params.mass_ratio,
                              dtype=tf.float32)

    prior_dict = GwPriors()._asdict()
    wvf_param_dict = waveform_params._asdict()
    detector_params_dict = detector_params._asdict()

    for param in waveform_params._fields:
        prior_val += prior_dict[param].log_prob(tf.cast(wvf_param_dict[param], tf.float32))

    for param in detector_params._fields:
        prior_val += prior_dict[param].log_prob(tf.cast(detector_params_dict[param], tf.float32))

    return tf.squeeze(prior_val)



"""
tensorflow compatible class representing a time series

builds upon gwastro/pycbc TimeSeries
https://github.com/gwastro/pycbc/blob/master/pycbc/types/timeseries.py

"""
import tensorflow as tf
import numpy as np
from scrinet.analysis import frequencyseries
from scrinet.analysis.tf_fft import fft


class TimeSeries(object):
    """Models a time series consisting of uniformly sampled scalar values.
    The time series can be complex but the FFTs only support real time
    series.

    Parameters
    ----------
    initial_array : array-like
        Array containing sampled data.
    delta_t : float
        Time between consecutive samples in seconds.
    epoch : {None, float}, optional
        Time of the first sample in seconds.
    dtype : {None, tensorflow or numpy data-type}, optional
        Sample data type.
    Attributes
    ----------
    data (tf.Tensor)
    delta_t
    duration
    start_time
    end_time
    sample_times
    sample_rate
    """

    def __init__(self, initial_array, delta_t, epoch=None, dtype=None, cast=tf.float64):
        self._name = "TimeSeries"
        if dtype is None:
            dtype = initial_array.dtype
        self._dtype = tf.dtypes.as_dtype(dtype)

        if self._dtype == tf.complex64:
            dt_dtype = tf.float32
        elif self._dtype == tf.complex128:
            dt_dtype = tf.float64
        else:
            # not sure about this catch
            dt_dtype = self._dtype

        self.delta_t = tf.cast(delta_t, dt_dtype)
        if epoch is None:
            epoch = 0
        self._epoch = tf.cast(epoch, dt_dtype)
        self.data = tf.convert_to_tensor(initial_array, self._dtype)

        self.sample_times = self.get_sample_times()
        self.duration = self.get_duration()
        self.sample_rate = self.get_sample_rate()

        self.start_time = self.get_start_time()
        self.end_time = self.get_end_time()

    def get_sample_times(self):
        if self._epoch is None:
            return range(self.data.shape[0]) * self.delta_t
        else:
            return range(self.data.shape[0]) * self.delta_t + self._epoch

    def get_duration(self):
        return self.data.shape[0] * self.delta_t

    def get_sample_rate(self):
        """Return the sample rate of the time series.
        """
        return tf.cast(tf.math.round(1.0 / self.delta_t), tf.int64)

    def get_start_time(self):
        """Return time series start time.
        """
        return self._epoch

    def get_end_time(self):
        """Return time series end time
        """
        return self._epoch + self.get_duration()

    def to_frequencyseries(self, delta_f=None):
        """ Return the Fourier transform of this time series
        Parameters
        ----------
        delta_f : {None, float}, optional
            The frequency resolution of the returned frequency series. By
        default the resolution is determined by the duration of the timeseries.
        Returns
        -------
        FrequencySeries:
            The fourier transform of this time series.
        """
        if self._dtype in [tf.complex64, tf.complex128]:
            raise NotImplementedError("only FFT of real time series supported")

        if not delta_f:
            delta_f = 1.0 / self.duration
        else:
            raise NotImplementedError("custom delta_f not supported")
            delta_f = tf.cast(delta_f, self.duration.dtype)

        # add 0.5 to round integer
        tlen = tf.cast(1.0 / delta_f / self.delta_t + 0.5, tf.int64)
        # flen = tf.cast(tlen / 2 + 1, tf.int64)

        # not sure how to assert this with symbolic tensor??
        # if tlen < self.data.shape[0]:
        #     raise ValueError("The value of delta_f (%s) would be "
        #                      "undersampled. Maximum delta_f "
        #                      "is %s." % (delta_f, 1.0 / self.duration))

        # zero pad the end of the data to the requied length
        # to get desired delta_f
        # tmp_t = np.zeros(tlen)
        # tmp_t[:self.data.shape[0]] = self.data.numpy()[:]
        # tmp_t = tf.convert_to_tensor(tmp_t, self._dtype)

        new_len = tf.math.abs(tlen - self.data.shape[0])
        arr_to_append = tf.zeros(new_len, dtype=self.data.dtype)
        tmp_t = tf.concat([self.data[:], arr_to_append], axis=0)
        tmp = TimeSeries(tmp_t, delta_t=self.delta_t,
                         epoch=self.start_time)
        fs = fft(tmp)
        return fs



import numpy as np
import os
import datetime

import tensorflow
from tensorflow.keras.models import Sequential
from tensorflow.keras.models import load_model
from tensorflow.keras.layers import Dense, Activation, Dropout
from tensorflow.keras.layers import LeakyReLU, ReLU, ELU
from tensorflow.keras.callbacks import ModelCheckpoint, TensorBoard, ReduceLROnPlateau, EarlyStopping
from tensorflow.keras import regularizers
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.optimizers.schedules import ExponentialDecay, InverseTimeDecay

from sklearn.preprocessing import StandardScaler

from . import scale


class RegressionANN(object):
    """
    Use tensorflow.keras to build an artifical neural network for regression
    """

    def __init__(self):
        pass

    def fit(self, X, y, input_dim, noutput, epochs,
            validation_data=None, outdir='./', outname='best.h5',
            scaleX=True, scaleY=True, verbose=0, batch_size=None,
            activation='relu',
            input_units=128,
            units=[128, 128, 128, 128, 128],
            learning_rate=0.001,
            lr_schedule=False,
            use_alr=False,
            alr_monitor='val_loss',
            alr_factor=0.1,
            alr_patience=5,
            alr_verbose=0,
            alr_min_lr=1e-4,
            use_es=False,
            es_monitor='val_loss',
            es_patience=10,
            es_verbose=0,
            kernel_initializer='glorot_uniform',
            use_tfboard=False,
            tfboard_logdir='logs/fit',
            tfboard_tag=None,
            tfboard_histogram_freq=0
            ):
        """
        input_dim: {int}. dimension of input layer
        noutput: {int}. number of outputs in final layer
        validation_data: (X_test, y_test)
        activation {str}: activation function to use for all layers
            (apart from last)
        input_units {int}: number of units in input layer
        units {list of ints}: number of units in each hidden layer.
            length of list is number of hidden layers
        use_alr {bool} default False.
            alr == adaptive learning rate
            if True then uses ReduceLROnPlateau
        alr_monitor {str, 'val_loss'}
        alr_factor {float, 0.1}
        alr_patience {int, 5}
        alr_verbose {int, 0 (or 1)}
        alr_min_lr {float, 1e-4}
        use_es {bool, False}
            es == early stopping
            if True then use EarlyStopping
        es_monitor{str, 'val_loss'}
        es_patience{int, 10}
        es_verbose{int, 0 (or 1)}
        kernel_initializer {str} default 'glorot_uniform'
        use_tfboard {bool: False} If true then enable TensorBoard logging
        tfboard_logdir {str: 'logs/fit'} path to log directory for TensorBoard
            callback.
        tfboard_tag {str: None}: identifier tag for TensorBoard sub directory
            If None then will use time stamp
        tfboard_histogram_freq {int: 0} histogram_freq in TensorBoard
        """
        self.outdir = outdir
        self.learning_rate = learning_rate
        self.use_alr = use_alr
        self.alr_monitor = alr_monitor
        self.alr_factor = alr_factor
        self.alr_patience = alr_patience
        self.alr_verbose = alr_verbose
        self.alr_min_lr = alr_min_lr
        self.use_es = use_es
        self.es_monitor = es_monitor
        self.es_patience = es_patience
        self.es_verbose = es_verbose
        self.kernel_initializer = kernel_initializer
        self.bestfile = os.path.join(self.outdir, outname)
        self.use_tfboard = use_tfboard
        self.tfboard_logdir = tfboard_logdir
        self.tfboard_histogram_freq = tfboard_histogram_freq

        if validation_data:
            X_test, y_test = validation_data

        self.scaleX = scaleX
        self.scaleY = scaleY
        if self.scaleX:
            X = X.copy()
            self.X_scalers = scale.make_scalers(X)
            X = scale.apply_scaler(X, self.X_scalers)
            if validation_data:
                X_test = scale.apply_scaler(X_test, self.X_scalers)
            self.save_X_scalers(os.path.join(self.outdir, "X_scalers"))
        if self.scaleY:
            y = y.copy()
            self.Y_scalers = scale.make_scalers(y)
            y = scale.apply_scaler(y, self.Y_scalers)
            if validation_data:
                y_test = scale.apply_scaler(y_test, self.Y_scalers)
            self.save_Y_scalers(os.path.join(self.outdir, "Y_scalers"))

        if validation_data:
            validation_data = (X_test, y_test)

        # Initialising the ANN
        self.model = Sequential()

        # Adding the input layer and the first hidden layer
        # self.model.add(Dense(128, kernel_regularizer=regularizers.l2(0.001), activation = 'relu', input_dim = input_dim))
        self.model.add(
            Dense(
                input_units,
                activation=activation,
                kernel_initializer=self.kernel_initializer,
                input_dim=input_dim
            )
        )
        # self.model.add(Dropout(0.2))

        for i in range(len(units)):
            self.model.add(
                Dense(units=units[i],
                      activation=activation,
                      kernel_initializer=self.kernel_initializer
                      ))

        # Adding the second hidden layer
        # self.model.add(Dense(units = 128, kernel_regularizer=regularizers.l2(0.001), activation = 'relu'))
        # self.model.add(Dense(units = 128, activation = 'relu'))
        # self.model.add(Dropout(0.2))

        # Adding the third hidden layer
        # self.model.add(Dense(units = 128, kernel_regularizer=regularizers.l2(0.001), activation = 'relu'))
        # self.model.add(Dense(units = 128, activation = 'relu'))
        # self.model.add(Dropout(0.2))

        # self.model.add(Dense(units = 128, kernel_regularizer=regularizers.l2(0.001), activation = 'relu'))
        # self.model.add(Dense(units = 128, activation = 'relu'))
        # self.model.add(Dropout(0.2))

        # self.model.add(Dense(units = 128, kernel_regularizer=regularizers.l2(0.001), activation = 'relu'))
        # self.model.add(Dense(units = 128, activation = 'relu'))
        # self.model.add(Dropout(0.2))

        # self.model.add(Dense(units = 128, kernel_regularizer=regularizers.l2(0.001), activation = 'relu'))
        # self.model.add(Dense(units = 128, activation = 'relu'))
        # self.model.add(Dropout(0.2))

        # Adding the output layer

        self.model.add(Dense(units=noutput))

        if lr_schedule:
            STEPS_PER_EPOCH = 1
            lr = InverseTimeDecay(
                0.001,
                decay_steps=STEPS_PER_EPOCH*500,
                decay_rate=1,
                staircase=False)
            # initial_learning_rate = 0.1
            # lr = ExponentialDecay(
            #     initial_learning_rate,
            #     decay_steps=100000,
            #     decay_rate=0.96,
            #     staircase=True)
        else:
            lr = self.learning_rate
        self.optimizer = Adam(learning_rate=lr)

        # Compiling the ANN
        self.model.compile(optimizer=self.optimizer,
                           loss='mean_squared_error', metrics=['mse'])

        if verbose:
            print(self.model.summary(), flush=True)

        callbacks = [ModelCheckpoint(
            f'{self.bestfile}', save_best_only=True, monitor='loss')]

        if self.use_tfboard:
            if tfboard_tag:
                self.tfboard_tag = tfboard_tag
            else:
                self.tfboard_tag = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            log_dir = os.path.join(self.tfboard_logdir, self.tfboard_tag)
            tensorboard_callback = TensorBoard(
                log_dir=log_dir, histogram_freq=self.tfboard_histogram_freq)
            callbacks.append(tensorboard_callback)

        if self.use_alr:
            reduce_lr = ReduceLROnPlateau(
                monitor=self.alr_monitor,
                factor=self.alr_factor,
                patience=self.alr_patience,
                min_lr=self.alr_min_lr,
                verbose=self.alr_verbose)
            callbacks.append(reduce_lr)

        if self.use_es:
            es = EarlyStopping(
                monitor=self.es_monitor,
                patience=self.es_patience,
                verbose=self.es_verbose
            )
            callbacks.append(es)

        # Fitting the ANN to the Training set
        history = self.model.fit(
            X,
            y,
            epochs=epochs,
            verbose=verbose,
            validation_data=validation_data,
            batch_size=batch_size,
            callbacks=callbacks
        )

        return history

    def load_model(self, file):
        self.model = load_model(file)

    def save_X_scalers(self, filename):
        np.save(filename, self.X_scalers)

    def save_Y_scalers(self, filename):
        np.save(filename, self.Y_scalers)

    def load_X_scalers(self, filename, allow_pickle=True):
        self.X_scalers = np.load(filename, allow_pickle=allow_pickle)

    def load_Y_scalers(self, filename, allow_pickle=True):
        self.Y_scalers = np.load(filename, allow_pickle=allow_pickle)

    def predict(self, X):
        if self.scaleX:
            X = scale.apply_scaler(X, self.X_scalers)

        y = self.model.predict(X)

        if self.scaleY:
            y = scale.apply_inverse_scaler(y, self.Y_scalers)

        return y



"""
tensorflow compatible class representing a frequency series
of a real time series

builds upon gwastro/pycbc FrequencySeries
https://github.com/gwastro/pycbc/blob/master/pycbc/types/frequencyseries.py

"""

import tensorflow as tf
import numpy as np
from scrinet.analysis import timeseries
from scrinet.analysis.tf_fft import ifft


def fftfreq(n, d):
    """
    Copy of https://numpy.org/doc/stable/reference/generated/numpy.fft.fftfreq.html

    NOTE: assumed casting to tf.float64 - should add optional dtype arg.

    Return the Discrete Fourier Transform sample frequencies.
    The returned float array `f` contains the frequency bin centers in cycles
    per unit of the sample spacing (with zero at the start).  For instance, if
    the sample spacing is in seconds, then the frequency unit is cycles/second.
    Given a window length `n` and a sample spacing `d`::
      f = [0, 1, ...,   n/2-1,     -n/2, ..., -1] / (d*n)   if n is even
      f = [0, 1, ..., (n-1)/2, -(n-1)/2, ..., -1] / (d*n)   if n is odd
    Parameters
    ----------
    n : int or float. It gets cast to tf.float64
        Window length.
    d : scalar.  It gets cast to tf.float64
        Sample spacing (inverse of the sampling rate).
    Returns
    -------
    f : Tensor
        An 1-D Tensor of type dtype
        Tensor of shape `(n,)` containing the sample frequencies.
    Examples
    --------
    >>> signal = tf.Variable([-2, 8, 6, 4, 1, 0, 3, 5], dtype=tf.complex128)
    >>> fourier = tf.signal.fft(signal)
    >>> n = signal.shape[0]
    >>> timestep = 0.1
    >>> freq = fftfreq(n, d=timestep)
    >>> freq
    tf.Tensor([ 0.          1.24999998  2.49999996  3.74999994 -4.99999993 -3.74999994 -2.49999996 -1.24999998], shape=(8,), dtype=float64)
    """
    n = tf.cast(n, tf.float64)
    d = tf.cast(d, tf.float64)
    val = 1.0 / (n * d)
    results = tf.zeros(tf.cast(n, tf.int64))
    N = (n-1)//2 + 1
    p1 = tf.range(0, N)
    p2 = tf.range(-(n//2), 0)
    results = tf.concat([p1, p2], axis=0)
    return results * val


class FrequencySeries(object):
    """Models a frequency series consisting of uniformly sampled scalar values.
    Parameters
    ----------
    initial_array : array-like
        Array containing sampled data.
    delta_f : float
        Frequency between consecutive samples in Hertz.
    epoch : {None, float}, optional
        Start time of the associated time domain data in seconds.
    dtype : {None, tensorflow or numpy data-type}, optional
        Sample data type.
    Attributes
    ----------
    delta_f : float
        Frequency spacing
    epoch : float
        Time at 0 index.
    sample_frequencies : Array
        Frequencies that each index corresponds to.
    """

    def __init__(self, initial_array, delta_f=None, epoch=None, dtype=None):
        self._name = "FrequencySeries"
        if dtype is None:
            dtype = initial_array.dtype

        self._dtype = tf.dtypes.as_dtype(dtype)

        if self._dtype == tf.complex64:
            df_dtype = tf.float32
        elif self._dtype == tf.complex128:
            df_dtype = tf.float64
        else:
            # not sure about this catch
            df_dtype = self._dtype

        self.delta_f = tf.cast(delta_f, df_dtype)
        if epoch is None:
            epoch = 0
        self._epoch = tf.cast(epoch, self._dtype)
        self.data = tf.convert_to_tensor(initial_array, self._dtype)

        self.sample_frequencies = self.get_sample_frequencies()

    def get_sample_frequencies(self):
        """Return an Array containing the sample frequencies.
        """
        return range(self.data.shape[0]) * self.delta_f

    def to_timeseries(self, delta_t=None):
        """ Return the Fourier transform of this time series.
        Note that this assumes even length time series!

        Parameters
        ----------
        delta_t : {None, float}, optional
            The time resolution of the returned series. By default the
        resolution is determined by length and delta_f of this frequency
        series.

        Returns
        -------
        TimeSeries:
            The inverse fourier transform of this frequency series.
        """
        nat_delta_t = 1.0 / \
            ((self.data.shape[0]-1)*2) / self.delta_f  # for real fft
        # nat_delta_t =  1.0 / ((len(self.data))) / self.delta_f  # for complex fft
        if not delta_t:
            delta_t = nat_delta_t
        else:
            raise NotImplementedError("custom delta_t not supported")
        delta_t = tf.cast(delta_t, self.delta_f.dtype)

        # add 0.5 to round integer
        tlen = tf.cast(1.0 / self.delta_f / delta_t + 0.5, tf.int64)
        flen = tf.cast(tlen / 2 + 1, tf.int64)  # for real fft
        # flen = tf.cast(tlen, tf.int64) # for complex fft

        # if flen < self.data.shape[0]:
        #     raise ValueError("The value of delta_t (%s) would be "
        #                      "undersampled. Maximum delta_t "
        #                      "is %s." % (delta_t, nat_delta_t))

        # tmp_f = np.zeros(tlen, dtype=np.complex128)
        # tmp_f[:self.data.shape[0]] = self.data.numpy()[:]
        # tmp_f = tf.convert_to_tensor(tmp_f, self._dtype)

        new_len = tf.math.abs(tlen-self.data.shape[0])
        arr_to_append = tf.zeros(new_len, dtype=self.data.dtype)
        tmp_f = tf.concat([self.data[:], arr_to_append], axis=0)
        tmp = FrequencySeries(tmp_f, delta_f=self.delta_f, epoch=self._epoch)
        ts = ifft(tmp)
        return ts



import numpy as np


def generate_degrees(maxdegs, max_deg_total=None):
    """
    generates an array of monomial exponents (or degrees)
    that can be used to generate a design matrix to perform
    linear basis polynomial regression

    intput:
        maxdegs: list of len number of dimensions
            list of max deg in each dimension
        max_deg_total: int (default: None)
            max degree of a term.
            If None then will use the maximum of `maxdegs`.

    returns:
        numpy.ndarray of shape (M, dim)
        M = total number of terms
        dim = dimension
    """
    if max_deg_total == None:
        max_deg_total = np.max(maxdegs)

    # create lists of degrees for each dimension
    xsdims = [range(deg+1) for deg in maxdegs]
    # tensor product them
    xs = np.meshgrid(*xsdims)
    # ravel them into a list into a 1D array
    xs = list(map(np.ravel, xs))
    # degrees is an np.ndarray of shape (M, dim)
    degrees = np.array(list(zip(*xs)))
    # restrict the max degree of each term to be less than max_deg_total+1
    mask_total_deg = np.sum(degrees, axis=-1) < max_deg_total + 1
    # restrict the max degree of each dimension
    mask_degs = [degrees[:, i] < maxdegs[i] + 1 for i in range(len(maxdegs))]
    # make a mega mask of all masks
    mask = np.prod(np.row_stack((mask_total_deg, *mask_degs)),
                   axis=0, dtype=bool)
    return np.array(degrees[mask])


class LinearPolynomialBasisRegression(object):
    """
    Linear Basis Function Regression with polynomials
    """

    def __init__(self):
        pass

    def make_degrees(self, maxdegs, max_deg_total):
        return generate_degrees(maxdegs, max_deg_total)

    def fit(self, X, y, maxdegs, method='direct', max_deg_total=None):
        """
        input:
            X: np.ndarray. shape (N, ndim)
            y: np.ndarray. shape (N, 1)
            method: str (default: 'direct')
                'direct': np.linalg.pinv. moore-penrose pseudo inverse
                'least-squres': np.linalg.lstsq. lapack least squares algorithm

            max_deg_total: int (default: None)
                max degree of a term.
                If None then will use the maximum of `maxdegs`.

        assigns the self.w_ml attribute.
        These are the maximum likelihood coefficients or weights.

        """

        self.degrees = generate_degrees(maxdegs, max_deg_total).tolist()

        Phi = self.make_design_matrix(X, self.degrees)

        if method == 'direct':
            self.w_ml = np.dot(np.linalg.pinv(Phi), y)
        elif method == 'least-squres':
            self.w_ml = np.linalg.lstsq(Phi, y, rcond=-1)[0]

    def make_design_matrix(self, X, degrees):
        Phi = np.stack([np.prod(X**d, axis=1) for d in degrees], axis=-1)
        return Phi

    def predict(self, X):

        Phi = self.make_design_matrix(X, self.degrees)
        return np.dot(Phi, self.w_ml)

    def loss(self, X, y, method='sum-of-squares', return_information_criteria=False):
        """
        input:
            X: np.ndarray.shape (N, ndim)
            y: np.ndarray.shape (N, 1)
            method:
                'sum-of-squres':
                'RMSE':
                'MSE':
            return_information_criteria: bool Default: False
                if True then returns AIC, AICc, BIC
        """

        yhat = self.predict(X)

        def mse(y, yhat):
            """
            mean-squared-error function
            """
            errors = (y - yhat)
            loss = np.sum(errors**2, axis=0) / len(y)
            return loss

        if method == 'sum-of-squares':
            errors = (y - yhat)
            loss = np.sum(errors**2, axis=0) / 2.0
        elif method == 'MSE':
            loss = mse(y, yhat)
        elif method == 'RMSE':
            loss = np.sqrt(mse(y, yhat))
        else:
            raise ValueError(f"method = {method} not valid.")

        if return_information_criteria:

            N = len(y)
            k = len(self.degrees) + 1
            likelihood = loss
            aic = self.AIC(N, likelihood, k)
            try:
                aicc = self.AICc(N, likelihood, k)
            except ZeroDivisionError:
                aicc = np.inf
            bic = self.BIC(N, likelihood, k)

            return loss, [aic, aicc, bic]
        else:
            return loss

    def AIC(self, N, likelihood, k):
        """
        N: number of samples
        likelihood: likelihood e.g. sum-of-squares
        k: degrees of freedom
        """
        aic = N * np.log(likelihood) + 2 * k
        return aic

    def AICc(self, N, likelihood, k):
        """
        N: number of samples
        likelihood: likelihood e.g. sum-of-squares
        k: degrees of freedom
        """
        aic = self.AIC(N, likelihood, k)
        aicc = aic + (2 * k * (k+1))/(N - k - 1)
        return aicc

    def BIC(self, N, likelihood, k):
        """
        N: number of samples
        likelihood: likelihood e.g. sum-of-squares
        k: degrees of freedom
        """
        bic = N*np.log(likelihood) + k*np.log(N)
        return bic


def findpolyfit(X, y, max_degs, output_best=False, validation_data=None, method='sum-of-squares', verbose=True):
    """
    X, y data to fit
    max_degs: max deg to try in each dim. e.g., if 2D X data then max_degs could be
        array([[0, 0],
            [1, 0],
            [2, 0],
            [0, 1],
            [1, 1],
            [0, 2]])
        In practice use fits.poly.generate_degrees([2,2])

    returns a bunch of stuff included the prefereed degree
    according to min([AIC, AICc, BIC])

    if output_best=True
    then you will simply get the best fit

    """
    if validation_data:
        Xval, y_val = validation_data
        # validation losses and info crit
        v_losses = []
        v_AICs = []
        v_AICcs = []
        v_BICs = []

    lrs = []
    # training and info crit
    t_losses = []
    t_AICs = []
    t_AICcs = []
    t_BICs = []

    for deg in max_degs:
        lr = LinearPolynomialBasisRegression()
        lr.fit(X, y, maxdegs=deg)
        tloss, [taic, taicc, tbic] = lr.loss(
            X, y, method=method, return_information_criteria=True)
        t_losses.append(tloss)
        t_AICs.append(taic)
        t_AICcs.append(taicc)
        t_BICs.append(tbic)

        min_t_AICs = np.argmin(t_AICs)
        min_t_AICcs = np.argmin(t_AICcs)
        min_t_BICs = np.argmin(t_BICs)

        lrs.append(lr)

        if validation_data:
            vloss, [vaic, vaicc, vbic] = lr.loss(
                Xval, y_val, method=method, return_information_criteria=True)
            v_losses.append(vloss)
            v_AICs.append(vaic)
            v_AICcs.append(vaicc)
            v_BICs.append(vbic)
            min_v_AICs = np.argmin(v_AICs)
            min_v_AICcs = np.argmin(v_AICcs)
            min_v_BICs = np.argmin(v_BICs)

    lowest_deg_t = min([min_t_AICs, min_t_AICcs, min_t_BICs])
    lowest_loss_t = np.min(t_losses)

    if verbose:
        print(f"lowest_deg for training set = {lowest_deg_t}")
        print(f"lowest loss for training set = {lowest_loss_t}")

    if validation_data:
        lowest_deg_v = min([min_v_AICs, min_v_AICcs, min_v_BICs])
        lowest_loss_v = np.min(v_losses)

        if verbose:
            print(f"lowest_deg for validation set = {lowest_deg_v}")
            print(f"lowest loss for validation set = {lowest_loss_v}")

    if output_best:
        return lrs[lowest_deg_t], lowest_deg_t

    if validation_data:
        return lrs[lowest_deg_t], lowest_deg_t, lrs, (lowest_deg_t, t_losses, t_AICs, t_AICcs, t_BICs), (lowest_deg_v, v_losses, v_AICs, v_AICcs, v_BICs)
    else:
        return lrs[lowest_deg_t], lowest_deg_t, lrs, (lowest_deg_t, t_losses, t_AICs, t_AICcs, t_BICs)
